//
// TX script execution with subcommand-handler callback and TxPrint routing
//
// Author: J. van Wijk
//
// Developed for LPT/DFS utilities
//

#include <txlib.h>                              // TX library interface
#include <txwpriv.h>                            // TX library private interface

#define TXSC_HELPREG        0xffff0220          // base help-id registration

#define TXSC_PRAGMA         ";;"
#define TXSC_P_PARDEF       ";;defaultparam"
#define TXSC_P_PARSET       ";;setparam"
#define TXSC_P_SISTEP       ";;singlestep"
#define TXSC_P_VERBOS       ";;verbose"
#define TXSC_P_PROMPT       ";;prompt"
#define TXSC_P_ERRORS       ";;error"

static  char       *runhelp[] =
{
   "",
   "Run a SCRIPT-file with a sequence of commands"
#if defined (DEV32)
   ", or containing REXX code",
#else
   " to be executed",
#endif
   "",
   "usage:       RUN  scriptfile-name  [options] [arguments]",
   "",
#if defined (DEV32)
   " Options:    (Options and Arguments for NATIVE scripts, not REXX)",
#else
   " Options:",
#endif
   "   -v      = verbose, show details about running the script",
   "   -p-     = no prompting and command-echo by the command handler",
   "   -p:1    = echo command by handler, but no prompting afterwards",
   "   -p:2    = no echo by handler, but prompting afterwards",
   "   -p:3    = echo before, and prompt after executing (default)",
   "   -E:i    = ignore errors, continue execution when RC is not 0",
   "   -E:q    = quit on any errors, exit  (default when in batch mode)",
   "   -E:c    = confirm by user on errors (default when interactive)",
   "   -s      = single-step, confirm each line before executing it",
   "   -q      = quit the program after executing the script",
   "",
   " Arguments : upto 9 arguments can be specified and can be used anywhere",
   "             in the script as '$1' to '$9' while $0 is the scriptname.",
   "             Arguments can be enclosed in single- or double-quotes",
   "             to allow embedded spaces or special characters.",
   "",
   NULL
};

#if defined (USEWINDOWING)
static BOOL txs_help_registred = FALSE;

static  char       *confirmhelp[] =
{
   "",
   " Script command execution prompt",
   "",
   " This is to acknowledge that you want the displayed command to",
   " be executed, taking the latest return-code (RC) into account.",
   "",
   "",
   "",
   "",
   "",
   NULL
};
#endif


// Substitute $n parameters in a string by supplied values; n = 0..9
static ULONG txsSubstituteParams                // RET   nr of substitutions
(
   char               *string,                  // IN    base string
   char              **values,                  // IN    values to substitute
   char               *result                   // OUT   substituted (MAXLIN)
);


/*****************************************************************************/
// Run native TXS script using callback (from parsed RUN cmd if runcmd == NULL)
/*****************************************************************************/
ULONG TxsNativeRun
(
   char               *name,                    // IN    TXS script to execute
   char               *runcmd,                  // IN    RUN cmd string or NULL
   TX_NATIVE_CALLBACK  subcom                   // IN    subcommand handler
)
{
   ULONG               rc = NO_ERROR;           // function return
   FILE               *script = NULL;
   TXLN                scrline;                 // script line
   TXLN                command;
   ULONG               linenr = 0;
   ULONG               scep = TxaOptNum('p', " prompt/echo ", 3);
   BOOL                loud = TxaOption('v');   // verbose script handling
   BOOL                step = TxaOption('s');   // single step
   BOOL                execute = TRUE;          // execute next command
   BOOL                abort   = FALSE;         // abort whole script
   char                error;                   // type of error handling
   BOOL                confirm;                 // confirm continuation
   TXTM                basename;                // script basename ($0)
   char               *params[10];              // script parameters
   char               *pardef[10];              // default parameter values
   char               *pp;                      // parameter
   char               *s;                       // generic char pointer
   int                 pi;                      // parameter index

   ENTER();

   #if defined (USEWINDOWING)
      if (!txs_help_registred)
      {
         txwRegisterHelpText( TXSC_HELPREG, "Script confirmation", confirmhelp);
         txs_help_registred = TRUE;
      }
   #endif                                       // USEWINDOWING

   error = TxaErrorStrategy( 'E', TxaExeSwitch('b'));

   if (subcom != NULL)                          // command handler present ?
   {
      if ((script = TxFindAndOpenFile( name, "PATH", scrline)) != NULL)
      {
         if (loud)                              // verbose ?
         {
            TxPrint( "\nRUN native script : '%s'\n", scrline);
         }
         strcpy( basename, TxGetBaseName( name));
         TxStripExtension( basename);

         for (pi = 0; pi < 10; pi++)            // params 0..9
         {
            pardef[ pi] = NULL;                 // default parameter value
            params[ pi] = TxaArgValue( pi +1);  // parsed arguments
         }
         params[0] = basename;                  // use base script name

         while (!abort && !TxAbort() && !feof(script) && !ferror(script))
         {
            strcpy( scrline, "");
            confirm = FALSE;
            if (fgets( scrline, TXMAXLN, script) != NULL)
            {
               linenr++;
               TxRepl( scrline, '\n', '\0');
               TxRepl( scrline, '\r', '\0');    // mainly for Linux/Unix ...
               txsSubstituteParams( scrline, params, command);

               if ( (strlen( command) > 0) &&   // not empty, and
                   ((command[0] != ';')    ||   // not a comment
                    (command[1] == ';')))       // except when a pragma
               {
                  if (step)                     // single step ?
                  {
                     confirm = TRUE;
                  }
                  else if (rc != NO_ERROR)      // some error
                  {
                     switch (error)
                     {
                        case 'q': abort   = TRUE;    break;
                        case 'c': confirm = TRUE;    break;
                        default:                     break;
                     }
                  }
                  if (!abort)
                  {
                     execute = TRUE;
                     if (confirm)
                     {
                        if (!TxConfirm( TXSC_HELPREG + 00,
                            "Script line nr : % 2lu,  Last RC was : %ld\n"
                            "Next command is:\n\n%s%s%s\n\n"
                            "Execute this command ? [Y/N] : ",
                             linenr, rc, CBC, command, CNN))
                        {
                           execute = FALSE;
                        }
                     }
                     if (execute)
                     {
                        if (loud)               // verbose ?
                        {
                           TxPrint( "RUN exec line % 3lu : '%s'\n",
                                     linenr, command);
                        }
                        if (strncmp( command, TXSC_PRAGMA,
                                       strlen(TXSC_PRAGMA)) == 0)
                        {
                           TRACES(("Pragma: '%s'\n", command));
                           for (pp = command; *pp && (*pp != ' '); pp++)
                           {
                           }
                           while (*pp && (*pp == ' '))
                           {
                              pp++;             // skip whitespace
                           }
                           TRACES(("Params: '%s'\n", pp));
                           if (strnicmp( command, TXSC_P_PARDEF,
                                          strlen( TXSC_P_PARDEF)) == 0)
                           {
                              if (isdigit(*pp))
                              {
                                 pi = (int) ((*pp) - '0');        //- index 0..9
                                 if ((       params[pi]  == NULL) ||
                                     (strlen(params[pi]) == 0))   //- need default ?
                                 {
                                    pp += 2;
                                    TxFreeMem( pardef[pi]);
                                    if ((pardef[pi] = TxAlloc(  1,  strlen(pp) + 1))
                                                   != NULL)
                                    {
                                       if (loud) // verbose ?
                                       {
                                          TxPrint( "RUN def  $%d value : '%s'\n", pi, pp);
                                       }
                                       strcpy( pardef[pi], pp);
                                       if ((s = strchr( pardef[pi], ';')) != NULL)
                                       {
                                          *s = 0; // strip comment from value
                                       }
                                       //- also removes translated CR/LF on Linux
                                       TxStrip( pardef[pi], pardef[pi], 0, ' ');
                                       params[pi] = pardef[pi];
                                    }
                                 }
                              }
                           }
                           else if (strnicmp( command, TXSC_P_PARSET,
                                               strlen( TXSC_P_PARSET)) == 0)
                           {
                              if (isdigit(*pp))
                              {
                                 TXLN        value;

                                 strcpy( value, params[pi]); //- current is default
                                 pi = (int) ((*pp) - '0');   //- index 0..9
                                 pp += 2;                    //- skip param number
                                 TxFreeMem( pardef[pi]);     //- free old value
                                 if (loud)                   //- verbose ?
                                 {
                                    TxPrint( "RUN set  $%d value : '%s'\n", pi, pp);
                                 }
                                 if (*pp == '?')             //- prompt for a value
                                 {
                                    pp += 2;                 //- skip questionmark

                                    TxPrompt( 0, TXMAXLN, value, pp);
                                    if ((pardef[pi] = TxAlloc(  1,
                                                      strlen(value) + 1)) != NULL)
                                    {
                                       strcpy(  pardef[pi], value);
                                       TxStrip( pardef[pi], pardef[pi], 0, ' ');
                                       params[pi] = pardef[pi];
                                    }
                                 }
                                 else           // direct set of a value
                                 {
                                    if ((pardef[pi] = TxAlloc(  1,
                                                         strlen(pp) + 1)) != NULL)
                                    {
                                       strcpy( pardef[pi], pp);
                                       if ((s = strchr( pardef[pi], ';')) != NULL)
                                       {
                                          *s = 0; // strip comment from value
                                       }
                                       //- also removes translated CR/LF on Linux
                                       TxStrip( pardef[pi], pardef[pi], 0, ' ');
                                       params[pi] = pardef[pi];
                                    }
                                 }
                              }
                           }
                           else if (strnicmp( command, TXSC_P_VERBOS,
                                                strlen(TXSC_P_VERBOS)) == 0)
                           {
                              BOOL      verbose = loud;

                              loud = TxaParseBool( pp);
                              if (loud || verbose)
                              {
                                 TxPrint( "RUN set verbose   : %s\n",
                                           (loud) ? "ON" : "OFF");
                              }
                           }
                           else if (strnicmp( command, TXSC_P_SISTEP,
                                                strlen(TXSC_P_SISTEP)) == 0)
                           {
                              step = TxaParseBool( pp);
                              if (loud)         // verbose ?
                              {
                                 TxPrint( "RUN set singlestep: %s\n",
                                           (step) ? "ON" : "OFF");
                              }
                           }
                           else if (strnicmp( command, TXSC_P_PROMPT,
                                                strlen(TXSC_P_PROMPT)) == 0)
                           {
                              if (isdigit(*pp))
                              {
                                 scep = (ULONG) (*pp - '0');
                                 if (loud)      // verbose ?
                                 {
                                    TxPrint( "RUN set prompting : %lu\n",
                                              scep);
                                 }
                              }
                           }
                           else if (strnicmp( command, TXSC_P_ERRORS,
                                                strlen(TXSC_P_ERRORS)) == 0)
                           {
                              if (strchr( "icqICQ", *pp) != NULL)
                              {
                                 error = tolower(*pp);
                                 if (loud)      // verbose ?
                                 {
                                    TxPrint( "RUN set error mode: '%c'\n",
                                              error);
                                 }
                              }
                           }
                           else                 // unrecognized pragma
                           {
                              if (loud)
                              {
                                 TxPrint( "RUN warning Pragma: '%s' unknown!\n",
                                           command);
                              }
                           }
                        }
                        else                    // execute by subcom handler
                        {
                           TRACES(("Subcom: '%s'\n", command))
                           rc = (subcom)( command,
                                          (scep & 0x01),
                                          (scep & 0x02),
                                           TRUE);
                        }
                     }
                  }
                  else
                  {
                     TxPrint( "Aborting execution\n");
                  }
               }
               else
               {
                  if (loud)                     // verbose ?
                  {
                     TxPrint( "RUN skip line % 3lu : '%s'\n",
                               linenr, command);
                  }
               }
            }
         }
         for (pi = 0; pi < 10; pi++)            // params 0..9
         {
            TxFreeMem( pardef[ pi]);            // free allocated defaults
         }
         fclose( script);
      }
      else
      {
         TxPrint( "TXs native script : '%s' not found\n", name);
         rc = TX_INVALID_FILE;
      }
   }
   else                                         // give usage for RUN
   {
      TxShowTxt( runhelp);
   }
   RETURN (rc);
}                                               // end 'TxsNativeRun'
/*---------------------------------------------------------------------------*/

/*****************************************************************************/
// Substitute $n parameters in a string by supplied values; n = 0..9
/*****************************************************************************/
static ULONG txsSubstituteParams                // RET   nr of substitutions
(
   char               *string,                  // IN    base string
   char              **values,                  // IN    values to substitute
   char               *result                   // OUT   substituted (MAXLIN)
)
{
   ULONG               rc = 0;                  // function return
   char               *bs = string;
   char               *rs = result;
   int                 rl = 0;                  // resulting length
   int                 si;                      // substitution index

   ENTER();
   TRARGS(("string: '%s'\n", string));

   while ((*bs) && (rl < TXMAXLN))
   {
      if ((*bs == '$') && isdigit(*(bs+1)))     // substitution candidate
      {
         bs++;                                  // skip  $
         si  = (int) ((*bs) - '0');             // index 0..9
         if (values[si] != NULL)                // value available ?
         {
            rl += strlen( values[si]);
            if (rl < TXMAXLN)                   // room to grow ?
            {
               strcpy(  rs,  values[si]);       // copy value
               rs += strlen( values[si]);       // and update destination
            }
         }
         bs++;                                  // skip digit
      }
      else
      {
         *rs++ = *bs++;                         // copy single char
         rl++;
      }
   }
   *rs = '\0';                                  // terminate result string

   TRARGS(("result: '%s'\n", result));
   RETURN (rc);
}                                               // end 'txsSubstituteParams'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Test if script exists, is REXX, return parameter description and full-name
/*****************************************************************************/
BOOL TxsValidateScript                          // RET   script exists
(
   char               *sname,                   // IN    script name
   BOOL               *rexx,                    // OUT   script is REXX
   char               *pdesc,                   // OUT   param description or NULL
   char               *fname                    // OUT   full scriptname or NULL
)
{
   BOOL                rc = FALSE;              // function return
   TXLN                line;
   FILE               *script;
   char               *pp = NULL;               // parameter position

   ENTER();

   if ((script = TxFindAndOpenFile( sname, "PATH", line)) != NULL)
   {
      rc = TRUE;                                // script exists
      if (fname != NULL)                        // full name wanted
      {
         strcpy( fname, line);                  // copy full name
         TRACES(("fname: '%s'\n", fname));
      }
      if (fgets( line, TXMAXLN, script) != NULL)
      {
         if (rexx != NULL)                      // want REXX test ?
         {
            *rexx = (memcmp( line, "/" "*", 2) == 0);
         }
         if (pdesc != NULL)                     // want descriptions ?
         {
            if (fgets( line, TXMAXLN, script) != NULL)
            {
               if ((pp = strrchr( line, '~')) != NULL)
               {
                  *(pp+1) = 0;                  // clean up the line
               }
               if ((pp = strchr( line, '~') ) != NULL)
               {
                  TxRepl( pp, '~', '\n');       // make descriptions readable
                  strcpy( pdesc, pp);           // and copy to caller ...
               }
               else
               {
                  strcpy( pdesc, "");
               }
            }
            TRACES(("pdesc: '%s'\n", pdesc));
         }
      }
      fclose( script);
   }
   BRETURN (rc);
}                                               // end 'TxsValidateScript'
/*---------------------------------------------------------------------------*/



#if defined (DEV32)                             // REXX for OS/2 only
#define TX_SIOEXIT  ((PSZ) "tx_sioexit")        // name of exit function

#define TX_REXXDLL  ((PSZ) "REXX")              // name of REXX dll
#define TX_REXXAPI  ((PSZ) "REXXAPI")           // name of REXXAPI dll

PFN txRexxStart;
PFN txRexxRegExe;
PFN txRexxRegExit;
PFN txRexxDeregEx;
PFN txRexxDereg;
PFN txRexxVarPool;

// Route 'say' output through the TxPrint channel
static LONG _System tx_sioexit
(
   LONG               exitnr,                   // IN    exit-id (RXSIO)
   LONG               funcnr,                   // IN    func-id
   PEXIT              parmblock                 // IN    parameters to func
);


/*****************************************************************************/
// Execute a REXX command script using a subcommand environment
/*****************************************************************************/
ULONG TxsRexxRun
(
   char               *name,                    // IN    REXX proc to execute
   char               *args,                    // IN    Arguments
   char               *envname,                 // IN    REXX environment name
   TX_SUBCOM_CALLBACK  subcom                   // IN    subcommand handler
)
{
   ULONG               dr = NO_ERROR;
   LONG                rc;                      /* rc from REXX interpreter  */
   RXSTRING            arg;                     /* argument string for REXX  */
   RXSTRING            rxret;                   /* return value from REXX    */
   SHORT               rexxrc = 0;              /* return code from function */
   TXLN                rxcode;
   HMODULE             rexxdll;                 // handle for REXX dll
   HMODULE             rexxapi;                 // handle for REXXAPI dll
   RXSYSEXIT           exit_list[2];            // list of exit-handlers

   ENTER();

   if (TxaOption('v'))                          // verbose ?
   {
      TxPrint( "\nRUN  REXX  script : '%s %s'\n", name, args);
   }
   dr = DosLoadModule( rxcode, TXMAXLN,         // error buffer and size
                       TX_REXXDLL,              // name of dll
                       &rexxdll);               // module handle
   if (dr == NO_ERROR)
   {
      dr = DosLoadModule( rxcode, TXMAXLN,      // error buffer and size
                          TX_REXXAPI,           // name of dll
                          &rexxapi);            // module handle
      if (dr == NO_ERROR)
      {
         TRACES(("REXX/API mod handles: %lu / %lu\n", rexxdll, rexxapi));
         if ((DosQueryProcAddr(rexxapi, 0,
             "RexxRegisterSubcomExe", (PFN *) &txRexxRegExe ) == NO_ERROR) &&
             (DosQueryProcAddr(rexxapi, 0,
             "RexxRegisterExitExe",   (PFN *) &txRexxRegExit) == NO_ERROR) &&
             (DosQueryProcAddr(rexxapi, 0,
             "RexxDeregisterExit",    (PFN *) &txRexxDeregEx) == NO_ERROR) &&
             (DosQueryProcAddr(rexxapi, 0,
             "RexxDeregisterSubcom",  (PFN *) &txRexxDereg  ) == NO_ERROR) &&
             (DosQueryProcAddr(rexxdll, 0,
             "RexxStart",             (PFN *) &txRexxStart  ) == NO_ERROR) &&
             (DosQueryProcAddr(rexxdll, 0,
             "RexxVariablePool",      (PFN *) &txRexxVarPool) == NO_ERROR)
            )
         {
            dr = txRexxRegExe( envname, subcom, NULL );
            TRACES(("RegSubcomExe RC: %lu\n", (ULONG) dr));
            if (dr == RXSUBCOM_OK)
            {
               dr = txRexxRegExit( TX_SIOEXIT, (PFN) &tx_sioexit, NULL );
               if (dr == RXSUBCOM_OK)
               {
                  exit_list[0].sysexit_name = TX_SIOEXIT;
                  exit_list[0].sysexit_code = RXSIO;
                  exit_list[1].sysexit_code = RXENDLST;

                  rxret.strlength = 0L;         /* force REXX to alloc rxret */
                  MAKERXSTRING(arg, args, strlen(args));

                  TRACES(("RexxStart to be executed\n"));
                  rc = txRexxStart(
                          1,                    /* number of arguments       */
                          &arg,                 /* array of arguments        */
                          (PSZ) name,           /* name of REXX file         */
                          0,                    /* No INSTORE used           */
                          envname,              /* Command env. name         */
                          RXSUBROUTINE,         /* invoked as subroutine     */
                          exit_list,            /* used EXITs on this call   */
                          &rexxrc,              /* Rexx program output       */
                          &rxret );             /* Rexx program output       */
                  if (rc == 0)
                  {
                     if (rexxrc != 0)
                     {
                        dr = (ULONG) rexxrc;
                     }
                  }
                  else
                  {
                     dr = (ULONG) 20000-rc;
                     TxPrint("Rexx interpreter error: %d on macro '%s'\n", rc, name);
                     sprintf( rxcode, "help REX%04.4u", (USHORT) (0-rc));
                     TxExternalCommand( rxcode);
                  }
                  if (rxret.strlength != 0)
                  {
                     //- to be refined, may want to pass that back
                     //- as a string to caller ?

                     TxPrint( "REXX return: '%*.*s'\n",
                               rxret.strlength, rxret.strlength, rxret.strptr);
                  }
                  DosFreeMem(rxret.strptr);     /* Release storage           */
                  txRexxDeregEx( TX_SIOEXIT, NULL);
               }
               else
               {
                  TxPrint("Error registering SAY exit handler\n");
               }
               txRexxDereg( envname, NULL);
            }
            else
            {
               TxPrint("Error registering REXX subcommand handler\n");
            }
         }
         else
         {
            TxPrint("Error resolving REXX API's from: '%s' and/or '%s'\n",
                     TX_REXXDLL, TX_REXXAPI);
         }
         DosFreeModule( rexxapi);
      }
      else
      {
         TxPrint("Error %lu loading Rexx API DLL: '%s'\n", dr, TX_REXXAPI);
      }
      DosFreeModule( rexxdll);
   }
   else
   {
      TxPrint("Error %lu loading Rexx DLL: '%s'\n", dr, TX_REXXDLL);
      TxPrint("Rexx functionality not available\n");
   }
   RETURN (dr);
}                                               // end 'TxsRexxRun'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Route 'say' output through the TxPrint channel
/*****************************************************************************/
static LONG _System tx_sioexit
(
   LONG                exitnr,                  // IN    exit-id (RXSIO)
   LONG                funcnr,                  // IN    func-id
   PEXIT               parmblock                // IN    parameters to func
)
{
   ULONG               dr;
   RXSIOSAY_PARM      *sio = (RXSIOSAY_PARM *) parmblock;

   ENTER();
   TRARGS(("Rexx exit: %ld, subfunction: %ld\n", exitnr, funcnr));

   switch (funcnr)
   {
      case RXSIOSAY:                            // normal Say output
      case RXSIOTRC:                            // trace and error output
         TxPrint( "%s\n", sio->rxsio_string.strptr);
         dr = RXEXIT_HANDLED;
         break;

      default:
         dr = RXEXIT_NOT_HANDLED;
         break;
   }
   RETURN (dr);
}                                               // end 'tx_sioexit'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Set number variable in the REXX variable pool using printf syntax
/*****************************************************************************/
ULONG TxsRexxSetnum
(
   char              *name,                     // IN    Name of variable
   ULONG              ulv,                      // IN    value for variable
   char              *fmt                       // IN    format strings
)
{
   ULONG              dr;
   TXLN               value;

   sprintf(value, fmt, ulv);
   TRACES(("Value to rx-set: '%s'\n", value));
   dr = TxsRexxSetvar( name, value, strlen(value));
   return (dr);
}                                               // end 'TxsRexxSetnum'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Set a string variable in the REXX variable pool
/*****************************************************************************/
ULONG TxsRexxSetvar
(
   char              *name,                     // IN    Name of variable
   void              *value,                    // IN    ptr to value buffer
   ULONG              length                    // IN    length of value
)
{
   ULONG               dr = 1;
   SHVBLOCK            var;                     // REXX variable pool block
   int                 i  = 9;                  // maximum 9 retries to avoid
                                                // RxVar bug when dynamic
                                                // loading is used
   ENTER();

   var.shvnext      = NULL;
   var.shvcode      = RXSHV_SYSET;
   var.shvnamelen   = strlen(name);
   var.shvvaluelen  = length;
   var.shvret       = 0;
   MAKERXSTRING(var.shvname,  name,  strlen(name));
   MAKERXSTRING(var.shvvalue, value, length);

   TRACES(("Set REXX variable: '%s', length: %lu, value:\n", name, length));
   TRHEXS(70, value, length, "REXX variable contents");
   while (i-- && ((dr = txRexxVarPool(&var)) != 0))
   {
      TRACES(("RexxVariablePool failed, retry!\n"));
   }
   TRACES(("RexxVariablePool rc: %lu, shvret: %lu\n", dr, var.shvret));
   RETURN (dr);
}                                               // end 'TxsRexxSetvar'
/*---------------------------------------------------------------------------*/

#endif
