// TX library selection-list manager
//
// Author: J. van Wijk
//
// JvW  26-06-2003 Initial version

#include <txlib.h>                              // public interface

/*****************************************************************************/
// Get length of longest string in a list
/*****************************************************************************/
ULONG TxSelGetMaxLength                         // RET   longest string length
(
   TXSELIST           *list                     // IN    selection list
)
{
   ULONG               rc = 0;                  // function return
   TXS_ITEM           *item;
   int                 l;

   ENTER();

   if (list != NULL)
   {
      for (l = 0; l < list->count; l++)
      {
         if ((item = list->items[l]) != NULL)
         {
            if (item->text != NULL)             // might be NULL (separator)
            {
               if (strlen( item->text) > rc)
               {
                  rc = strlen( item->text);
               }
               TRACES(("Item %2u len:%2u '%s'\n", l, strlen(item->text), item->text));
            }
            else
            {
               TRACES(("Item %2u len: 0  Probably a separator\n", l));
            }
         }
      }
   }
   RETURN (rc);
}                                               // end 'TxSelGetMaxLength'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Scroll visible list up by one line (towards lower index)
/*****************************************************************************/
ULONG TxSelScrollUp                             // RET   scrolling succeeded
(
   TXSELIST           *list                     // INOUT selection list
)
{
   ULONG               rc = NO_ERROR;              // function return

   ENTER();

   if ((list->vsize > 1) && (list->top != 0))   // scroll-up possible ?
   {
      if (list->renderNewItem)                  // render-new available ?
      {
         ULONG     index = list->vsize -1;      // index of discarded item
         TXS_ITEM *temp  = list->items[index];  // keep item allocation

         txSelItemDestroy( list->astatus & TXS_AS_DSTRINGS, &temp);

         memmove( &(list->items[1]),            // destination of move
                  &(list->items[0]),            // source of move
                   (list->vsize -1) * sizeof(TXS_ITEM *));

         list->items[0] = temp;                 // re-use same item struct
         rc = (list->renderNewItem)( list->items[0],
                                     NULL, list->top -1);
      }
      list->top--;                              // update visible top position
   }
   else
   {
      rc = TX_INVALID_DATA;
   }
   RETURN (rc);
}                                               // end 'TxSelScrollUp'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Scroll visible list down by one line (towards higher index)
/*****************************************************************************/
ULONG TxSelScrollDown                           // RET   scrolling succeeded
(
   TXSELIST           *list                     // INOUT selection list
)
{
   ULONG               rc = NO_ERROR;           // function return

   ENTER();

   if (             (list->vsize  > 1) &&
       ((list->top + list->vsize) < list->count)) // scroll-down possible ?
   {
      if (list->renderNewItem)                  // render-new available ?
      {
         ULONG         index = list->vsize -1;  // index of new item
         TXS_ITEM     *temp  = list->items[0];  // keep item allocation

         txSelItemDestroy( list->astatus & TXS_AS_DSTRINGS, &temp);

         memmove( &(list->items[0]),            // destination of move
                  &(list->items[1]),            // source of move
                   (list->vsize -1) * sizeof(TXS_ITEM *));

         list->items[index] = temp;             // re-use same item struct
         rc = (list->renderNewItem)( list->items[index],
                                     NULL, list->top + index +1);
      }
      list->top++;                              // update visible top position
   }
   else
   {
      rc = TX_INVALID_DATA;
   }
   RETURN (rc);
}                                               // end 'TxSelScrollDown'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Set visible list to start at specified external index; Discard old items
/*****************************************************************************/
ULONG TxSelSetPosition                          // RET   resulting top index
(
   TXSELIST           *list,                    // INOUT selection list
   ULONG               index                    // IN    top external index
)
{
   ULONG               newtop = index;          // function return

   ENTER();

   if ((newtop + list->vsize) > list->count)    // set top beyond end list ?
   {
      if (list->count > list->vsize)
      {
         newtop = list->count - list->vsize;    // set to end
      }
      else
      {
         newtop = 0;                            // end equals beginning ...
      }
   }
   if (list->renderNewItem)                     // render-new available ?
   {
      ULONG         i;

      for (i = 0; i < list->vsize; i++)
      {
         TXS_ITEM     *item  = list->items[i];  // keep item allocation

         if (item == NULL)                      // not allocated yet
         {
            list->astatus |= TXS_AS_DYNITEMS;
            item = TxAlloc( 1, sizeof( TXS_ITEM));
         }
         else                                   // discard string values
         {
            txSelItemDestroy( list->astatus & TXS_AS_DSTRINGS, &item);
         }
         if (item != NULL)
         {
            (list->renderNewItem)( item, NULL, list->top + i);
         }
         list->items[i] = item;
      }
   }
   list->top = newtop;                          // update visible top position
   RETURN (newtop);
}                                               // end 'TxSelSetPosition'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Set selection for specified index, and/or bits if multiple select possible
/*****************************************************************************/
ULONG TxSelSetSelected                          // RET   resultcode
(
   TXSELIST           *list,                    // INOUT selection list
   ULONG               index,                   // IN    external index
   TXSETRESET          action,                  // IN    set, reset or toggle
   BYTE                mask                     // IN    bits to set/reset
)
{
   ULONG               rc = NO_ERROR;           // function return

   ENTER();

   if (index < list->tsize)
   {
      if (list->selarray != NULL)               // multiple select OK ?
      {
         BYTE  old = list->selarray[index];
         BYTE  new;

         switch (action)
         {
            case TX_SET:   new = old |  mask; break;
            case TX_RESET: new = old & ~mask; break;
            default:       new = old ^  mask; break;
         }
         if ((old == 0) && (new != 0) && (list->selected < list->tsize))
         {
            list->selected++;
         }
         if ((old != 0) && (new == 0) && (list->selected != 0))
         {
            list->selected--;
         }
         list->selarray[index] = new;
      }
      else
      {
         list->selected = (action == TX_RESET) ? 0 : index;
      }
   }
   else
   {
      rc = TX_INVALID_DATA;
   }
   RETURN (rc);
}                                               // end 'TxSelSetSelected'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Itterate over selected list items, calback all with specified mask set
/*****************************************************************************/
ULONG TxSelItterateSelected
(
   TXSELIST           *list,                    // INOUT selection list
   TXS_ITEMHANDLER     selected,                // IN    callback selected
   BYTE                mask                     // IN    selection mask
)
{
   ULONG               rc = NO_ERROR;           // function return
   TXS_ITEM            item;

   ENTER();

   if (list->selarray != NULL)                  // multiple select OK ?
   {
      if (list->selected != 0)
      {
         ULONG            i;

         for (i = 0; i < list->tsize; i++)      // visit all list entries
         {
            if ((list->selarray[i] & mask) == mask) // matches selection ?
            {
               if (list->vsize == list->tsize)  // all info available
               {
                  (selected)(list->items[i], &(list->selarray[i]), i);
               }
               else if (list->renderNewItem)
               {
                  (list->renderNewItem)(&item, &(list->selarray[i]), i);
                  (selected)(&item, NULL, i);
               }
               else
               {
                  (selected)( NULL, NULL, list->selected);
               }
            }
         }
      }
      else                                      // nothing selected
      {
         rc = TX_INVALID_DATA;
      }
   }
   else                                         // callback for selected
   {
      if (list->vsize == list->tsize)           // all info available
      {
         (selected)(list->items[list->selected], NULL, list->selected);
      }
      else if (list->renderNewItem)
      {
         (list->renderNewItem)(&item, NULL, list->selected);
         (selected)(&item, NULL, list->selected);
      }
      else
      {
         (selected)( NULL, NULL, list->selected);
      }
   }
   RETURN (rc);
}                                               // end 'TxSelItterateSelected'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Set 'selected' to item with specified selection char (at text[index-1])
/*****************************************************************************/
ULONG TxSelCharSelect                           // RET   index now selected
(
   TXSELIST           *list,                    // INOUT selection list
   char                select                   // IN    selection character
)
{
   ULONG               rc = list->selected;
   TXS_ITEM           *item;

   ENTER();
   TRACES(("list:%8.8lx select: '%c'\n", list, select));

   if (list->selarray == NULL)                  // single-select only
   {
      ULONG            i;

      for (i = 0; i < list->count; i++)         // visit list entries
      {
         if ((item = list->items[i]) != NULL)
         {
            if (item->index != 0)               // selection char there ?
            {
               if (list->items[i]->text != NULL) // has text string ?
               {
                  if (item->text[ item->index -1] == select)
                  {
                     list->selected = i;        // set new selection
                     rc = i;
                     break;                     // end the search
                  }
               }
            }
         }
      }
   }
   RETURN (rc);
}                                               // end 'TxSelCharSelect'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Set 'selected' to item where text starts with 'string' (case-insensitive)
/*****************************************************************************/
ULONG TxSelStringSelect                         // RET   index now selected
(
   TXSELIST           *list,                    // INOUT selection list
   char               *select,                  // IN    selection string
   int                 length                   // IN    significant length
)
{
   ULONG               rc = 0;
   TXS_ITEM           *item;

   ENTER();
   TRACES(( "list:%8.8lx  select:'%s'  length:%u\n", list, select, length));

   if (list != NULL)
   {
      rc = list->selected;
      TRACES(("list:%8.8lx select: '%s'\n", list, select));

      if (list->selarray == NULL)               // single-select only
      {
         ULONG            i;

         for (i = 0; i < list->count; i++)      // visit list entries
         {
            if ((item = list->items[i]) != NULL) // existing item ?
            {
               if (list->items[i]->text != NULL) // has text string ?
               {
                  if (strnicmp( list->items[i]->text, select, length) == 0)
                  {
                     list->selected = i;        // set new selection
                     rc = i;
                     break;                     // end the search
                  }
               }
            }
         }
      }
   }
   RETURN (rc);
}                                               // end 'TxSelStringSelect'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Get unique quick-select character for specified list index
/*****************************************************************************/
USHORT TxSelGetCharSelect                       // RET   index for select-char
(                                               //       or 0 if not possible
   TXSELIST           *list,                    // IN    selection list
   ULONG               nr,                      // IN    item to get char for
   int                 offset                   // IN    minimum position 0..n
)
{
   USHORT              rc = 0;                  // function return
   TXS_ITEM           *item;
   TXLN                used;                    // used characters
   int                 l;
   char                c;

   ENTER();

   memset( used, 0, TXMAXLN);

   if (list != NULL)
   {
      //- build list of used characters

      for (l = 0; l < list->count; l++)
      {
         if (l != nr)                           // can replace by same value!
         {
            if ((item = list->items[l]) != NULL)
            {
               if ((item->text != NULL) && (item->index != 0))
               {
                  used[ strlen(used)] = item->text[item->index -1];
               }
            }
         }
      }
      strupr( used);                            // convert to upper-case!
      TRACES(("item-nr %lu, Selchars used: '%s'\n", nr, used));

      if ((item = list->items[nr]) != NULL)     // item to test
      {
         for (l = offset; l < strlen( item->text); l++)
         {
            c = toupper( item->text[l]);
            if (isalnum(c))                     // accept alpha-numeric only
            {
               if (strchr( used, c) == NULL)    // not used yet
               {
                  rc = l +1;
                  TRACES(("SelChar: '%c' at index %hu\n", item->text[l], rc));
                  break;
               }
            }
         }
      }
   }
   RETURN (rc);
}                                               // end 'TxSelGetCharSelect'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Set/reset specified flag-bit(s) in menu-item by searching menu-bar structure
/*****************************************************************************/
ULONG TxSetMenuItemFlag                         // RET   #items found and set
(
   TXS_MENUBAR        *mbar,                    // IN    menubar information
   ULONG               mid,                     // IN    value, menu item id
   BYTE                mask,                    // IN    bits to set/reset
   BOOL                set                      // IN    set specified bit(s)
)
{
   ULONG               rc = 0;                  // function return
   TXS_MENU           *menu;
   int                 m;

   ENTER();
   TRACES(("%sSET item: %lu, using mask %2.2hx\n", (set) ? "" : "RE", mid, mask));

   for (m = 0; m < mbar->count; m++)
   {
      if ((menu = mbar->menu[m]) != NULL)
      {
         TRACES(("Menu %d='%s' list:%8.8lx\n", m +1, menu->text, menu->list));

         rc += TxSetListItemFlag( 1, menu->list, mid, mask, set);
      }
   }
   RETURN (rc);
}                                               // end 'TxSetMenuItemFlag'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Set/reset specified flag-bit(s) in menu-item by searching list-structure
/*****************************************************************************/
ULONG TxSetListItemFlag                         // RET   #items found and set
(
   ULONG               recurse,                 // IN    recursion level (1)
   TXSELIST           *list,                    // IN    list information
   ULONG               mid,                     // IN    value, menu item id
   BYTE                mask,                    // IN    bits to set/reset
   BOOL                set                      // IN    set specified bit(s)
)
{
   ULONG               rc = 0;                  // function return
   TXS_ITEM           *item;
   int                 i;

   TRACES(("%sSET list item: %lu, using mask %2.2hx on list at %8.8lx\n",
          (set) ? "" : "RE", mid, mask, list));

   if (list != NULL)
   {
      for (i = 0; i < list->count; i++)
      {
         // TRACES(("Check item: %u\n", i));
         if ((item = list->items[i]) != NULL)   // item to test
         {
            if (item->value == mid)             // matching item id ?
            {
               if (set)
               {
                  item->flags |=  mask;
               }
               else
               {
                  item->flags &= ~mask;
               }
               rc++;                            // count item as set

               /*
               TRACES(("List %8.8lx item %d='%s' %sSET %s %s\n",
                        list, i +1, item->text, (set) ? "" : "RE",
                        (mask & TXSF_DISABLED) ? "DISABLED" : "",
                        (mask & TXSF_MARKED  ) ? "MARKED"   : ""));
               */
            }
            if ((item->flags & TXSF_P_LISTBOX) && // this is a popup/submenu
                (item->user != 0))              // and a list is attached.
            {
               // TRACES(( "Recurse level %lu list: %8.8lx\n", recurse +1, item->user));

               if (recurse < 99)                // sanity check
               {
                  rc += TxSetListItemFlag( recurse +1,
                                          (TXSELIST *) item->user,
                                           mid, mask, set);
               }
               else
               {
                  TRACES(("Aborting on insane recursion level ...\n"));
               }
            }
         }
      }
   }
   return (rc);
}                                               // end 'TxSetListItemFlag'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Create a dynamic selection list with specified size
/*****************************************************************************/
ULONG TxSelCreate
(
   ULONG               vsize,                   // IN    visible size
   ULONG               tsize,                   // IN    total size of list
   ULONG               asize,                   // IN    size for item array
   ULONG               astat,                   // IN    alloc status for items
   BOOL                multi,                   // IN    create multi-select info
   TXS_ITEMHANDLER     render,                  // IN    callback render new
   TXSELIST          **list                     // OUT   selection list
)
{
   ULONG               rc = NO_ERROR;           // function return
   TXSELIST           *selist;

   ENTER();
   TRACES(("vsize:%hu  tsize:%hu  asize:%hu  astat:%8.8lx\n",
            vsize, tsize, asize, astat));

   if ((selist = TxAlloc( 1, sizeof(TXSELIST))) != NULL)
   {
      selist->astatus       = astat | TXS_AS_LISTRUCT;
      selist->renderNewItem = render;
      selist->asize         = asize;
      selist->vsize         = vsize;
      selist->tsize         = tsize;
      selist->count         = 0;                // no items yet

      if (asize != NULL)
      {
         TRACES(( "Allocate item-ptr array for %lu items\n", tsize));
         if ((selist->items  = TxAlloc( asize, sizeof(TXS_ITEM *))) != NULL)
         {
            selist->astatus |= TXS_AS_PTRARRAY;
         }
         else
         {
            rc = TX_ALLOC_ERROR;
         }
      }
      if (multi && (tsize != 0))
      {
         TRACES(( "Allocate selection array for %lu items\n", tsize));
         if ((selist->selarray = TxAlloc( tsize, sizeof(BYTE))) != NULL)
         {
            selist->astatus   |= TXS_AS_SELARRAY;
         }
         else
         {
            rc = TX_ALLOC_ERROR;
         }
      }
      *list = selist;                           // return created list
      if (rc != NO_ERROR)
      {
         txSelDestroy( list);                   // free partial list memory
      }
   }
   else
   {
      rc = TX_ALLOC_ERROR;
   }
   TRACES(("list:%8.8lx\n", *list));
   RETURN (rc);
}                                               // end 'TxSelCreate'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Destroy dynamic parts of the selection list, freeing memory
/*****************************************************************************/
ULONG txSelDestroy
(
   TXSELIST          **list                     // IN    selection list
)
{
   ULONG               rc = NO_ERROR;           // function return

   ENTER();

   if ((list != NULL) && ((*list) != NULL))
   {
      ULONG            as = (*list)->astatus;   // allocation status
      ULONG            ic = (*list)->count;     // actual number of items
      ULONG            i;

      if ((as & TXS_AS_DYNITEMS) != 0)          // dynamic alloc in items ?
      {
         for (i = 0; i < ic; i++)
         {
            txSelItemDestroy( as, &((*list)->items[i]));
         }
      }
      if (as & TXS_AS_PTRARRAY)
      {
         TRACES(("Destroying item ptr-array at:%8.8lx\n", (*list)->items));
         TxFreeMem( (*list)->items);            // free item array
      }
      if (as & TXS_AS_SELARRAY)
      {
         TRACES(("Destroying multi selarray at:%8.8lx\n", (*list)->selarray));
         TxFreeMem( (*list)->selarray);         // free selection array
      }
      if (as & TXS_AS_LISTRUCT)
      {
         TRACES(("Destroying list at:%8.8lx\n", *list));
         TxFreeMem( (*list));                   // free list structure
      }
   }
   else                                         // invalid parameters
   {
      rc = TX_INVALID_DATA;
   }
   RETURN (rc);
}                                               // end 'txSelDestroy'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Destroy dynamic parts of a selection list-item, freeing memory
/*****************************************************************************/
ULONG txSelItemDestroy
(
   ULONG               as,                      // IN    allocaton status
   TXS_ITEM          **itemref                  // IN    selection list item
)
{
   ULONG               rc = NO_ERROR;           // function return

   ENTER();

   if ((itemref != NULL) && ((*itemref) != NULL))
   {
      TXS_ITEM  *item = *itemref;

      if (as & TXS_AS_SEL_DESC)
      {
         TxFreeMem( item->desc);
      }
      if (as & TXS_AS_SEL_TEXT)
      {
         TxFreeMem( item->text);
      }
      if (as & TXS_AS_SELITEMS)
      {
         TxFreeMem( (*itemref));
      }
   }
   else                                         // invalid parameters
   {
      rc = TX_INVALID_DATA;
   }
   RETURN (rc);
}                                               // end 'txSelItemDestroy'
/*---------------------------------------------------------------------------*/


/*****************************************************************************/
// Create selist from existing Txt list
/*****************************************************************************/
ULONG TxSelistFromTxt                           // RET   result
(
   char               *txt[],                   // IN    NULL terminated txt
   BOOL                multi,                   // IN    keep multi-select info
   BOOL                share,                   // IN    share data, no alloc
   TXSELIST          **list                     // OUT   selection list
)
{
   ULONG               rc = NO_ERROR;           // function return
   ULONG               items = 0;

   ENTER();

   if (txt != NULL)
   {
      while (txt[items] != NULL)
      {
         items++;                               // count lines in txt
      }
      rc = TxSelCreate( items, items, items,
                        TXS_AS_DYNBASED | TXS_AS_SELITEMS,
                        multi, NULL, list);

      if (rc == NO_ERROR)
      {
         TXSELIST     *selist = *list;
         ULONG         line;

         if (!share)                            // allocated text strings ?
         {
            selist->astatus |= TXS_AS_SEL_TEXT;
         }
         for (line = 0; (line < items) && (rc == NO_ERROR); line++)
         {
            TXS_ITEM  *newitem = TxAlloc( 1, sizeof(TXS_ITEM));

            if (newitem != NULL)
            {
               if (share)                       // just copy string pointer
               {
                  newitem->text = txt[line];
               }
               else
               {
                  if ((newitem->text = TxAlloc(1,strlen(txt[line]) +1)) != NULL)
                  {
                     strcpy( newitem->text, txt[line]);
                  }
                  else
                  {
                     rc = TX_ALLOC_ERROR;
                  }
               }
               selist->items[line] = newitem;
               selist->count       = line +1;   // number of items in list
            }
            else
            {
               rc = TX_ALLOC_ERROR;
            }
         }
         if (rc != NO_ERROR)
         {
            txSelDestroy( list);                // free partial list memory
         }
      }
   }
   else
   {
      rc = TX_INVALID_DATA;
   }
   RETURN (rc);
}                                               // end 'TxSelistFromTxt'
/*---------------------------------------------------------------------------*/

/*****************************************************************************/
// Create selist from existing selist (duplicate)
/*****************************************************************************/
ULONG TxSelistDuplicate                         // RET   result
(
   TXSELIST           *org,                     // IN    original Selist
   BOOL                multi,                   // IN    keep multi-select info
   BOOL                share,                   // IN    share data, no alloc
   TXSELIST          **list                     // OUT   selection list
)
{
   ULONG               rc = NO_ERROR;           // function return
   ULONG               items = 0;

   ENTER();

   if (org != NULL)
   {
      if (!share)
      {
         items = org->asize;
      }
      rc = TxSelCreate( org->vsize, org->tsize, items, 0,
                       (org->selarray != NULL), NULL, list);

      if (rc == NO_ERROR)
      {
         TXSELIST     *selist = *list;
         ULONG         line;

         if (share)
         {
            selist->items = org->items;         // just point to same info
            selist->count = org->count;
         }
         else
         {
            selist->astatus |= TXS_AS_DYNITEMS;
            for (line = 0; (line < items) && (rc == NO_ERROR); line++)
            {
               TXS_ITEM  *orgitem = org->items[line];
               TXS_ITEM  *newitem = TxAlloc( 1, sizeof(TXS_ITEM));

               if (newitem != NULL)
               {
                  if (((newitem->text = TxAlloc(1,strlen(orgitem->text) +1)) != NULL) &&
                      ((newitem->desc = TxAlloc(1,strlen(orgitem->desc) +1)) != NULL)  )
                  {
                     strcpy( newitem->text, orgitem->text);
                     strcpy( newitem->desc, orgitem->desc);
                  }
                  else
                  {
                     rc = TX_ALLOC_ERROR;
                  }
                  selist->items[line] = newitem;
                  selist->count       = line +1; // number of items in list
               }
               else
               {
                  rc = TX_ALLOC_ERROR;
               }
            }
         }
         if (rc == NO_ERROR)
         {
            if (multi && (org->selarray))       // keep same multi-selected
            {                                   // items selected
               memcpy( selist->selarray, org->selarray, org->tsize);
            }
         }
         else
         {
            txSelDestroy( list);                // free partial list memory
         }
      }
   }
   else
   {
      rc = TX_INVALID_DATA;
   }
   RETURN (rc);
}                                               // end 'TxSelistDuplicate'
/*---------------------------------------------------------------------------*/

/*****************************************************************************/
// Create single-item list with ONE disabled text-item (for 'empty' selections)
/*****************************************************************************/
TXSELIST *TxSelEmptyList                        // RET   selection list or NULL
(
   char               *text,                    // IN    text for single item
   char               *desc                     // IN    description (footer)
)
{
   TXSELIST           *list  = NULL;            // total list
   TXS_ITEM           *item;                    // single item

   ENTER();

   if (TxSelCreate( 1, 1, 1, TXS_AS_NOSTATIC, FALSE, NULL, &list) == NO_ERROR)
   {
      list->astatus  = TXS_AS_NOSTATIC;         // all dynamic allocated

      if ((item  = TxAlloc( 1, sizeof(TXS_ITEM))) != NULL)
      {
         list->count    = 1;                    // actual items in list
         list->items[0] = item;                 // attach item to list

         item->helpid = TXWH_USE_OWNER_HELP;    // from owner-menu-item

         if (((item->text = TxAlloc( 1, strlen(text) +1)) != NULL) &&
             ((item->desc = TxAlloc( 1, strlen(desc) +1)) != NULL)  )
         {
            item->flags   = TXSF_DISABLED;

            strcpy( item->text, text);
            strcpy( item->desc, desc);
         }
      }
   }
   RETURN( list);
}                                               // end 'TxSelEmptyList'
/*---------------------------------------------------------------------------*/

